import os
from shutil import rmtree
from typing import List, Optional
import typer
import docker

from skapp.models.stack import Stack, load_plugins
from skapp.utils import load_yaml_files

app = typer.Typer()


@app.command()
def version():
    pass


def parse_values(value):
    obj = {}

    for val in value:
        if "=" in val:
            key, v = val.split("=", 1)
        else:
            key, v = val, True

        if "." in key:
            _parent = obj
            while "." in key:
                el, key = key.split(".", 1)
                if el not in _parent:
                    _parent[el] = {}
                _parent = _parent[el]

            _parent[key] = v

        else:
            obj[key] = v

    return obj


@app.command()
def rollout(
    name: str,
    configs: Optional[List[str]] = typer.Option(None),
    value: Optional[List[str]] = typer.Option(None),
    namespace: str = typer.Option(None),
    create_namespace: bool = typer.Option(False),
    environment: Optional[str] = typer.Option("dev"),
    build: bool = typer.Option(False),
) -> None:
    values = parse_values(value)
    stack = Stack.from_files(
        name=name, environment=environment, files=configs, values=values
    )
    if build:
        build_images(stack)
    stack.chart.rollout(namespace=namespace, create_namespace=create_namespace)


@app.command()
def generate_chart(
    name: str,
    location: str,
    configs: Optional[List[str]] = typer.Option(None),
    namespace: str = typer.Option(None),
    environment: Optional[str] = typer.Option("dev"),
    build: bool = typer.Option(False),
) -> None:
    stack = Stack.from_files(name=name, environment=environment, files=configs)
    if build:
        build_images(stack)

    if os.path.exists(location):
        os.removedirs(location)

    os.makedirs(location)
    stack.chart.dump(
        location,
        namespace=namespace,
    )


def get_image_deps(img, images):
    deps = []
    for dep in img.dependencies:
        if dep in images:
            deps.append(images[dep])
            deps.extend(get_image_deps(images[dep], images))

    return deps


def build_images(stack: Stack, image: Optional[str] = None):
    if len(stack.build) == 0:
        return

    import datetime

    image_version = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
    client = docker.from_env()

    built_images = []

    images = {_.name: _ for _ in stack.build}

    while True:
        for img in images.values():
            if img.name in built_images:
                continue

            if image is not None and img.name != image:
                continue

            print(f"Building {img.name}")

            deps = get_image_deps(img, images)

            if not set(built_images).issuperset({_.name for _ in deps}):
                # print(f"Dependencies for {img.name} are not yet ready")
                continue

            img.build(client=client, image_version=image_version, deps=deps)

            built_images.append(img.name)

        if len(images) == len(built_images):
            break


@app.command()
def build(
    name: str,
    image: str = None,
    configs: Optional[List[str]] = typer.Option(None),
    value: Optional[List[str]] = typer.Option(None),
    namespace: str = typer.Option(None),
    environment: Optional[str] = typer.Option("dev"),
) -> None:
    values = parse_values(value)
    stack = Stack.from_files(
        name=name, environment=environment, files=configs, values=values
    )
    build_images(stack, image=image)


if __name__ == "__main__":
    app()



from argparse import Namespace
import os
from re import sub
from urllib.parse import urlparse

import requests
import yaml
from jinja2 import BaseLoader, Environment


def uri_validator(x):
    try:
        result = urlparse(x)
        return all([result.scheme, result.netloc])
    except Exception:
        return False


def merge(dict1, dict2):
    for k in set(dict1.keys()).union(dict2.keys()):
        if k in dict1 and k in dict2:
            if isinstance(dict1[k], dict) and isinstance(dict2[k], dict):
                yield (k, dict(merge(dict1[k], dict2[k])))
            else:
                # If one of the values is not a dict, you can't merge it.
                # Value from second dict overrides one in first, then we
                # move on.
                yield (k, dict2[k])
                # Alternatively, replace this with exception raiser to alert
                # you of value conflicts
        elif k in dict1:
            yield (k, dict1[k])
        else:
            yield (k, dict2[k])


def parse_yaml(content):
    result = {}
    for partial in yaml.safe_load_all(content):
        if not partial:
            continue

        include = partial.pop("include", [])
        included = {}
        for child in include:
            child_dict = dict(load_yaml_files(child))
            included = dict(merge(included, child_dict))

        if include:
            partial = dict(merge(included, partial))

        result = dict(merge(result, partial))

    return {k: v for k, v in result.items() if not k.startswith(".")}


def load_yaml_files(*args):
    def load_yaml_file(filepath) -> str:
        if uri_validator(filepath):
            return requests.get(filepath).content

        with open(filepath) as f:
            return f.read()

    def _load_all_files():
        for filepath in args:
            yield load_yaml_file(filepath)

    return parse_yaml("\n---\n".join(_ for _ in _load_all_files() if _))


def camelize(key) -> str:
    """camelCase given key"""
    if key in ["shared_preload_libraries"]:
        return key

    if "_" not in key:
        return key
    enumerated = enumerate(key.lower().split("_"))
    return "".join(_ if i == 0 else _.capitalize() for i, _ in enumerated)


def snakelize(s):
    return "_".join(
        sub(
            "([A-Z][a-z]+)", r" \1", sub("([A-Z]+)", r" \1", s.replace("-", " "))
        ).split()
    ).lower()


def dict_to_yaml(data, context: dict = None):
    def _format(res):
        if isinstance(res, dict):
            new = {}
            for k, v in res.items():
                k = camelize(k)
                if v is not None:
                    new[k] = _format(v)
            return new

        if isinstance(res, list):
            return [_format(_) for _ in res]

        if isinstance(res, LazyString):
            return res.render(context)

        if isinstance(res, str) and "{{" in res and "}}" in res:
            return LazyString(res).render(context)

        return res

    return yaml.safe_dump(_format(data))


class Context(object):
    def __init__(self, context):
        self.context = context

    def resolve(self, name):
        return name

    def prefix(self, name):
        return f"{self.context['stack'].name}-{name}"

    def fqdn(self, name):
        return self.prefix(name + ".{{- .Release.Namespace -}}.svc.cluster.local")

    def helm(self, template):
        return "{{ " + template + " }}"

    def envvar(self, name):
        return os.environ.get(name)


class LazyString(str):

    context: object = None

    def get_template(self):
        return self

    def render(self, context):
        rtemplate = Environment(loader=BaseLoader).from_string(self.get_template())
        images = {_.name: _.tag for _ in context["stack"].build}
        context = Context(context)

        return rtemplate.render(
            resolve=context.resolve,
            prefix=context.prefix,
            fqdn=context.fqdn,
            helm=context.helm,
            envvar=context.envvar,
            images=images,
            **context.context)


class Prefixed(LazyString):
    def get_template(self):
        return "{{ prefix('%s') }}" % self


class Fqdn(LazyString):
    def get_template(self):
        return "{{ fqdn(prefix('%s')) }}" % self



# Handler for CNPG Cluster
# apiVersion: postgresql.cnpg.io/v1
# kind: Cluster
# metadata:
#   name: main-db
#   namespace: default
# spec:
#   instances: 1

#   storage:
#     pvcTemplate:
#       accessModes:
#         - ReadWriteOnce
#       resources:
#         requests:
#           storage: 1Gi
#       storageClassName: local-storage


#   monitoring:
#     enablePodMonitor: true

from typing import Generator
from skapp.models.resource import NamespacedResource
from skapp.utils import merge


API_VERSION = 'postgresql.cnpg.io/v1'

# Sample input:
'''
thirdparty:cnpg:cluster:
    main-db:
        instances: 1
        storage:
            class: local-storage
            size: 1Gi
            access_mode: ReadWriteOnce
'''


class Cluster(NamespacedResource):
    api_version = API_VERSION
    _kind = 'Cluster'

    def generate(self, namespace: str = None) -> Generator:
        yield {}

class Clusters(ResourceList):
    pass

def handler(elements: dict[str, dict]) -> dict:
    patch = {}
    for elname, config in elements.items():
        cluster = Cluster(name=elname, **config)
        generated = list(cluster.generate())

        patch = dict(merge(patch, {"custom": generated}))



import os
import subprocess
import tempfile
from typing import Optional
from pydantic import BaseModel

# import docker
from python_on_whales import docker


class ImageBuild(BaseModel):
    prefix: str = "docker.io"
    push: bool = True
    pull: bool = False
    name: str
    context: str = "./"
    dockerfile: str = "Dockerfile"
    tag: Optional[str] = None
    image: Optional[str] = None
    platforms: list[str] = ["linux/arm64" if "arm64" in os.uname().machine else "linux/amd64"]
    target: Optional[str] = None
    md5version: Optional[list[str]] = []
    dependencies: Optional[list[str]] = []


    def get_md5version(self):
        import hashlib

        m = hashlib.md5()

        content = []
        for verfile in self.md5version:
            with open(verfile, "r") as f:
                content.append(f.read())

        m.update("".join(content).encode("utf-8"))
        return m.hexdigest()


    def build(self, client: "docker.DockerClient", image_version: str, deps: list["ImageBuild"] = []) -> str:
        # Use docker-py to build the image
        separator = ":" if ":" not in self.prefix else "-"

        if self.md5version:
            image_version = self.get_md5version()

        self.tag = f"{self.prefix}{separator}{self.name}-{image_version}"

        exists = False
        if self.md5version:
            try:
                for platform in self.platforms:
                    client.images.pull(self.tag, platform=platform)
            except:
                exists = False
            else:
                exists = True


        if exists:
            print(f"Image {self.tag} already exists")
        else:
            print(f"Building {self.name} as {self.tag} with dockerfile {self.dockerfile} in {self.context}")

            dockerfile_path = self.dockerfile
            if self.dependencies:
                # Generate tempfile
                dockerfile_path = tempfile.mktemp()

                with open(self.dockerfile, "r") as f:
                    content = f.read()
                    print(deps, self.dependencies)
                    content = content.format(**{dep.name: dep.tag for dep in deps})
                    with open(dockerfile_path, "w") as f:
                        f.write(content)


            docker_command = [
                "docker",
                "buildx",
                "build",
                "--platform",
                ",".join(self.platforms),
                "-t",
                self.tag,
                self.context,
                "-f",
                dockerfile_path,
                "--push" if self.push else "--load",
            ]

            if self.target is not None:
                docker_command.extend(["--target", self.target])

            print(" ".join(docker_command))
            process = subprocess.Popen(
                docker_command,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
            )

            # Stream and print the output
            for line in process.stdout:
                print(line, end="")

        # # return tag
        # # Wait for the command to finish
        # # process.wait()

        # # config = docker.buildx.bake(targets=self.platforms or ['linux/amd64'], load=True, push=self.push)

        # # breakpoint()

        # image = client.images.get(tag)

        # # image, logs = client.images.build(
        # #     path=self.context,
        # #     dockerfile=self.dockerfile,
        # #     tag=tag,
        # #     platform="linux/amd64",
        # # )
        # # for log in logs:
        # #     print(log)

        # # if self.push:
        # #     for pushlog in client.images.push(tag, stream=True):
        # #         print(pushlog)

        # self.tag = tag
        # self.image = image

    @classmethod
    def new(cls, **config) -> "ImageBuild":
        return cls(**config)


class Builds(list[ImageBuild]):
    item_class = ImageBuild



import importlib
import os
from distutils.version import LooseVersion, StrictVersion
from enum import Enum
from subprocess import STDOUT, CalledProcessError, check_output
from tempfile import TemporaryDirectory
from typing import List, Optional, Generator

import kubernetes
import yaml
from kubernetes import client, config
from pydantic import validator

from skapp.models.base import Base, YamlMixin
from skapp.models.pod_controllers import (
    CronjobList,
    DaemonsetList,
    DeploymentList,
    IngressList,
    StatefulsetList,
    VolumesList,
    JobList,
)
from skapp.models.build import Builds
from skapp.models.resource import Resource, ResourceList, ConfigMapList, ConfigMap
from skapp.utils import dict_to_yaml, load_yaml_files, parse_yaml, merge


def load_plugin(name):
    name = name.replace(':', '.')
    return importlib.import_module(f"{name}")


def load_plugins(defn):
    patch = {}
    to_pop = []
    for k, conf in defn.items():
        if ':' in k:
            # Couldn't pop while iterating
            to_pop.append(k)

            plugin = load_plugin(k)

            patch = dict(merge(patch, plugin.handler(conf)))

    for k in to_pop:
        defn.pop(k)

    if patch:
        defn = dict(merge(defn, patch))

    return defn


class Stack(Base, YamlMixin):
    class Config:
        arbitrary_types_allowed = True

    name: str
    environment: Optional[str]
    description: Optional[str]
    version: Optional[StrictVersion]
    app_version: Optional[LooseVersion]
    deployments: Optional[DeploymentList] = []
    daemonsets: Optional[DaemonsetList] = []
    statefulsets: Optional[StatefulsetList] = []
    cronjobs: Optional[CronjobList] = []
    jobs: Optional[JobList] = []
    configs: Optional[ConfigMapList] = []
    volumes: Optional[VolumesList] = []
    ingress: Optional[IngressList] = []
    values: Optional[list[dict[str, str]]] = []
    custom: Optional[dict[str, dict]] = {}
    build: Optional[Builds] = {}

    @staticmethod
    def inject_defaults(defn):
        defaults = defn.pop("defaults", {})
        if defaults:
            for rt in ["deployments", "daemonsets", "statefulsets", "cronjobs", "jobs"]:
                for r in defn.get(rt, {}).values():
                    if r.get('.nodefaults'):
                        continue

                    for k, dv in defaults.items():
                        # Add if not present
                        if k not in r:
                            r[k] = dv
                            continue

                        # Merge only if it's a dict
                        if isinstance(r[k], dict):
                            r[k] = dict(merge(dv, r[k]))
        return defn

    @staticmethod
    def new(name: str, definition: str,  environment: Optional[str] = "dev") -> "Stack":
        defn = load_plugins(parse_yaml(definition))
        defn = Stack.inject_defaults(defn)
        return Stack(name=name, environment=environment, **defn)

    @staticmethod
    def from_files(
        name: str,
        files: List[str],
        environment: Optional[str] = "dev",
        values: list[dict[str, str]] = {}
    ) -> "Stack":
        defn = dict(merge(load_yaml_files(*files), values))
        defn = Stack.inject_defaults(defn)
        defn = load_plugins(defn)

        return Stack(name=name, environment=environment, **defn)

    @property
    def chart(self):
        return HelmChart(stack=self, template_generator=self.yaml_files)

    def to_yaml(self, context: Optional[dict] = None) -> str:
        return super(Stack, self).to_yaml(
            context={"stack": self} if context is None else context
        )

    @property
    def get_all_resources(self):
        for k in self.__annotations__.keys():
            x = getattr(self, k)
            if isinstance(x, Resource):
                yield x
            if isinstance(x, ResourceList):
                for res in x:
                    yield res

    def generate(self, namespace: str = None):
        for k in self.__annotations__.keys():
            x = getattr(self, k)
            if hasattr(x, "generate"):
                for res in x.generate(namespace=namespace):
                    # breakpoint()
                    yield res
            elif k == "custom":
                for name, config in x.items():
                    config['metadata']['name'] = f"{self.name}-{config['metadata']['name']}"
                    yield config
                    # foo = dict_to_yaml(config, context={"stack": self})
                    # breakpoint()
                    # yield foo

    @validator(
        "deployments",
        "daemonsets",
        "statefulsets",
        "configs",
        "cronjobs",
        "jobs",
        "ingress",
        "build",
        pre=True,
        allow_reuse=True,
        check_fields=False
    )
    def validate_stack_fields(cls, value, field):
        if isinstance(value, list):
            return field.type_(value)

        if isinstance(value, dict):
            dl = field.type_()
            for name, config in value.items():
                if isinstance(config, field.type_.item_class):
                    dl.append(config)
                else:
                    dl.append(field.type_.item_class.new(name=name, **config))

            return dl

        raise ValueError(f"Invalid value for {field.type_.__name__}")


class HelmChart(Base, YamlMixin):
    class TypeEnum(str, Enum):
        application = "application"
        library = "library"

    stack: Stack
    type: TypeEnum = TypeEnum.application
    description: Optional[str] = "A Helm chart for Kubernetes created dynamically"

    def generate(self, namespace: str = None):
        return self.stack.generate(namespace=namespace)

    def generate_files(self, namespace: str = None) -> Generator:
        yield self.generate_info_file()
        for filename, template in self.yaml_files(
            namespace=namespace,
            context={
                "stack": self.stack,
                "namespace": namespace
            }
        ):
            yield f"templates/{filename}", template

    def generate_info_file(self):
        return "Chart.yaml", dict_to_yaml(
            {
                "name": self.stack.name,
                "description": self.stack.description,
                "version": self.stack.version or "0.1.0",
                "app_version": self.stack.app_version or self.stack.version or "0.1.0",
            },
            context=None,
        )

    def dump(self, folder, namespace: str = None) -> None:
        for filename, content in self.generate_files(namespace=namespace):
            absolute_filename = os.path.join(folder, filename)
            if "/" in filename:
                file_folder = absolute_filename.rsplit("/", 1)[0]
                os.makedirs(file_folder, exist_ok=True)

            with open(absolute_filename, "w") as f:
                f.write(content)

    def check_compatibility(self):
        server = client.VersionApi().get_code()
        library_version = StrictVersion(kubernetes.__version__)
        assert abs(int(server.minor) - library_version.version[0]) <= 1, (
            f"kubernetes library {library_version} is not "
            f"compatible with server version {server.git_version}"
        )

    def rollout(self, location: Optional[str] = None, namespace: str = None, create_namespace=False) -> "HelmRelease":
        # Load kubernetes config
        config.load_kube_config()

        self.check_compatibility()

        def _rollout(loc: str) -> HelmRelease:
            self.dump(loc, namespace=namespace)
            self.exec(self.rollout_command(loc, namespace=namespace or None, create_namespace=create_namespace))

        if location:
            return _rollout(location)

        with TemporaryDirectory() as location:
            return _rollout(location)

    def rollout_command(self, location, namespace=None, create_namespace=False):
        cmd = ["helm", "upgrade", "--install", "--force", self.stack.name, location]

        if namespace:
            cmd += ["--namespace", namespace]

        if create_namespace:
            cmd += ["--create-namespace"]

        return cmd

    def uninstall(self):
        return self.exec(["helm", "delete", self.stack.name])

    def exec(self, command: str):
        def _log(output):
            for line in output.decode("utf-8").splitlines():
                print(line)

        try:
            _log(check_output(command, stderr=STDOUT))
        except CalledProcessError as e:
            _log(e.output)
            raise e

    def get_kubernetes_resources(self):
        for res in self.stack.get_all_resources:
            yield res.get_from_kubernetes(stack=self.stack)


class HelmRelease(Base):
    resource_definitions: dict

    @classmethod
    def load(cls, name) -> "HelmRelease":
        command = ["helm", "get", "manifest", name]
        try:
            yaml_resources = check_output(command, stderr=STDOUT)
        except CalledProcessError as e:
            print(e.output)
            raise e

        return HelmRelease(
            resource_definitions={
                f"{res['kind']}/{res['metadata']['name']}": res
                for res in yaml.safe_load_all(yaml_resources)
            }
        )

    def refresh(self):
        for name, resource in self.resource_definitions.keys():
            pass
